(* Unification efficace sur les types rcursifs avec ranges. *)

(* Voici la structure de nos types. *)

type structure =
  | Variable
  | Int
  | Arrow of node * node
  | Record of node
  | RowCons of string * node * node
  | RowAbs
  | Pre of node
  | Abs

(* Voici nos noeuds. Un noeud est essentiellement une variable appartenant  une
   multi-quation. Nous utilisons le module [UnionFind] pour effectuer une
   association efficace entre variables et structure; une valeur de type
   [structure] reprsente l'unique type construit $\hat\tau$ de la multi-quation,
   s'il en existe un, et est gal  [Variable] sinon. *)

and node =
  structure UnionFind.point

(* Cette fonction cre une variable frache. *)

let variable () =
  UnionFind.fresh Variable

(* Ces fonctions construisent de nouveaux termes  partir de sous-termes existants. *)

let int () =
  UnionFind.fresh Int

let arrow node1 node2 =
  UnionFind.fresh (Arrow (node1, node2))

let record node =
  UnionFind.fresh (Record node)

let rowcons label node1 node2 =
  UnionFind.fresh (RowCons (label, node1, node2))

let rowabs () =
  UnionFind.fresh RowAbs

let pre node =
  UnionFind.fresh (Pre node)

let abs () =
  UnionFind.fresh Abs

(* Cette fonction effectue l'unification de deux noeuds, c'est--dire la rsolution
   d'une quation  rsoudre. Notons que la fonction ne renvoie aucun rsultat; elle
   se contente de modifier les structures existant en mmoire. Elle lve une exception
   en cas d'chec. *)

exception UnificationError of node * node

let rec unify node1 node2 =
  
  (* Si les noeuds appartiennent  la mme multi-quation, il n'y a rien  faire. *)

  if not (UnionFind.equivalent node1 node2) then begin

    (* Dterminons la structure de chacune des deux multi-quations. *)

    let struc1 = UnionFind.describe node1
    and struc2 = UnionFind.describe node2 in

    (* Fusionnons immdiatement les deux multi-quations. Il est important de le
       faire avant d'effectuer un appel rcursif  [unify], afin de garantir la
       terminaison. 

       Le sens de la fusion est ici choisi arbitrairement. *)

    UnionFind.alias node1 node2;

    (* Raisonnons par cas sur la structure des deux multi-quations et dduisons-en
       ventuellement de nouvelles quations  rsoudre. *)

    match struc1, struc2 with

    (* Si l'une est une variable, c'est--dire ne contient aucun type construit,
       il n'y a rien  faire. *)

    | Variable, _
    | _, Variable ->
	()

    (* Si les deux sont des types construits ayant le mme symbole de tte, on
       dcompose. *)

    | Int, Int ->
	()
    | Arrow (dom1, rng1), Arrow (dom2, rng2) ->
	unify dom1 dom2;
	unify rng1 rng2
    | Record content1, Record content2 ->
	unify content1 content2

    | Pre field1, Pre field2 ->
	unify field1 field2
    | Abs, Abs ->
	()

    (* Restent les cas des ranges. *)

    | RowCons (label1, head1, tail1), RowCons (label2, head2, tail2) ->

	(* Si les deux labels concident, c'est facile. *)

	if label1 = label2 then begin
	  unify head1 head2;
	  unify tail1 tail2
	end
	    
	(* Sinon, il faut appliquer l'quation de commutation. Noter qu'une
	   variable frache est ncessaire. *)
	       
	else begin
	  let tail = variable () in
	  unify tail1 (rowcons label2 head2 tail);
	  unify tail2 (rowcons label1 head1 tail)
	end

    | RowCons (_, head, tail), RowAbs
    | RowAbs, RowCons (_, head, tail) ->

	unify head (abs());
	unify tail (rowabs())

    | RowAbs, RowAbs ->
	()

    (* Les autres cas sont des erreurs, soit de sortage, soit d'unification.
       Pour les distinguer, il faudrait ajouter du code ci-dessous; nous
       ferons l'hypothse que l'erreur de sortage est impossible par construction,
       et nous signalons donc une erreur d'unification. *)

    | _, _ ->
	raise (UnificationError (node1, node2))

  end

(* Pour tester: *)

let _ =
unify (rowcons "x" (pre (int())) (rowcons "y" (pre (int())) (rowabs())))
      (rowcons "y" (pre (int())) (rowcons "x" (pre (variable())) (variable())))
;; (* doit russir *)

let _ =
unify (rowcons "x" (pre (int())) (rowabs()))
      (rowcons "y" (pre (int())) (rowcons "x" (pre (variable())) (variable())))
;; (* doit chouer *)

